import java.io.StringReader;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Array;
import java.net.URLEncoder;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

/**
   Class of utility methods.  For the most part these have been
   stolen from Utility classes in the JHymn project.
   
   @author Tom Austin
 */
public class ThaUtil
{
   /**
      Converts a String representation of a hex number to bytes.
    */
   public static byte[] hexStringToBytes(String s)
   {
     if (s == null)
        return null;

     byte[]  b = new byte[s.length() / 2];

     for (int i = 0; i < b.length; ++i)
        b[i] = (byte) to_int(s.substring(i * 2, i * 2 + 2), 0, 16);

     return b;
   }
   
   public static int to_int(String s)
   {
     return to_int(s, 0, 10);
   }
   
   public static int to_int(String s, int errorVal)
   {
      return to_int(s, errorVal, 10);
   }

   private static int to_int(String s, int errorVal, int radix)
   {
     if (s == null || s.length() == 0)
       return errorVal;

     if (s.charAt(0) == '+')
       s = s.substring(1);

     try {
       return Integer.parseInt(s, radix);
     }
     catch (NumberFormatException e) {
       return errorVal;
     }
   }
   
   public static byte[] fromBase64String(String s)
   {
     if (s == null)
       return null;
     else
       return fromBase64String(s.getBytes());
   }

   private static byte[] fromBase64String(byte[] b)
   {
     if (b == null || b.length == 0)
       return b;

      int     len = b.length;

     while (len > 0 && b[len - 1] == '=')
       --len;

      int     dlen = (len / 4) * 3; // Decoded length of string.

     if (len % 4 > 1)
       dlen += (len % 4) - 1;

     byte[]  tb = new byte[len];
     byte[]  db = new byte[dlen + 2];
     int     c;
     int     b0, b1, b2, b3;

     for (int i = 0; i < len; ++i) {
       c = b[i] & 0xFF;

       if      (65 <= c && c <=  90) // A-Z
         c -= 65;
       else if (97 <= c && c <= 122) // a-z
         c -= 71;
       else if (48 <= c && c <=  57) // 0-9
         c += 4;
       else if (c == 43)             // +
         c = 62;
       else if (c == 47)             // /
         c = 63;
       else
         c = 0;

       tb[i] = (byte) c;
     }

     for (int i = 0, j = 0; i < len; i += 4, j += 3) {
       b0 = tb[i];
       b1 = (i < len - 1 ? tb[i + 1] : 0);
       b2 = (i < len - 2 ? tb[i + 2] : 0);
       b3 = (i < len - 3 ? tb[i + 3] : 0);

       db[j    ] = (byte) (0xFF & ((b0 << 2) | (b1 >> 4)));
       db[j + 1] = (byte) (0xFF & ((b1 << 4) | (b2 >> 2)));
       db[j + 2] = (byte) (0xFF & ((b2 << 6) | b3));
     }

     db = resizeByteArray(db, dlen);

     return db;
   }
   
   public static String bytesToString(byte[] b)
   {
     return bytesToString(b, "");
   }

   private static String bytesToString(byte[] b, String separator)
   {
     if (b == null)
       return "";

     StringBuffer  sb = new StringBuffer();

     for (int i = 0; i < b.length; ++i) {
       if (i > 0)
         sb.append(separator);

       sb.append(toHexString(b[i], 2));
     }

     return sb.toString();
   }
   
   /**
      Encodes a string in the format needed for a url.
    */
   public static String urlEncode(String s)
   {
     try {
       return URLEncoder.encode(s, "UTF-8");
     }
     catch (UnsupportedEncodingException e) {
       return "?";
     }
   }
   
   /**
      Parse a string and return an XML document object.
    */
   public static Document getDocumentFromString(String s)
   {
      try {
         DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
         DocumentBuilder parser = factory.newDocumentBuilder();
         InputSource source = new InputSource(new StringReader(s));
         Document document = parser.parse(source);

         return document;
      }
      catch (Exception e) {
         return null;
      }
   }
   
   /**
      Return the value for the specified key.
      iTunes responses are often pairs of <key></key> <string></string> elements.
    */
   public static String getValueForKey(Node node, String key)
   {
     Node  match = findNodeWithText(node, "key", key);

     if (match == null)
       return null;

     Node  valueNode = findNextNonTextSibling(match);

     if (valueNode == null)
       return null;

     String  valueTag = valueNode.getNodeName();
     String  value;

     if ("false".equals(valueTag))
       value = "false";
     else if ("true".equals(valueTag))
       value = "true";
     else {
       value = getEnclosedText(valueNode);
       value = (value != null ? value : "");
     }

     return value;
   }
   
   private static Node findNodeWithText(Node node, String name, String text)
   {
     NodeList  children = node.getChildNodes();
     Node      child;
     Node      match;

     for (int i = 0; i < children.getLength(); ++i) {
       child = children.item(i);

       if (name.equals(child.getNodeName()) && text.equals(getEnclosedText(child)))
         return child;
       else if ((match = findNodeWithText(child, name, text)) != null)
         return match;
     }

     return null;
   }
   
   private static String getEnclosedText(Node node)
   {
     NodeList  children = node.getChildNodes();
     Node      child;

     for (int i = 0; i < children.getLength(); ++i) {
       child = children.item(i);

       if (child.getNodeType() == Node.TEXT_NODE)
         return child.getNodeValue();
     }

     return "";
   }

   private static Node findNextNonTextSibling(Node node)
   {
     while ((node = node.getNextSibling()) != null) {
       if (node.getNodeType() != Node.TEXT_NODE)
         return node;
     }

     return null;
   }
   
   public static String toHexString(int n, int digits)
   {
     return right(padLeft(Integer.toHexString(n), '0', digits),
           digits).toUpperCase();
   }

   public static String toHexString(long n, int digits)
   {
     return right(padLeft(Long.toHexString(n), '0', digits),
           digits).toUpperCase();
   }
   
   public static byte[] resizeByteArray(byte[] a, int newSize)
   {
     return (byte[]) resizeArray(a, newSize);
   }
   
   private static Object resizeArray(Object a, int newSize)
   {
     if (a == null)
       return null;
     else if (!a.getClass().isArray())
       throw new ArrayStoreException();

     int     oldSize = Array.getLength(a);

     if (oldSize == newSize)
       return a;

     Object    a2 = Array.newInstance(a.getClass().getComponentType(), newSize);

     System.arraycopy(a, 0, a2, 0, Math.min(oldSize, newSize));

     return a2;
   }

   private static String repeat(char c, int count)
   {
      if (count < 0)
         return null;
      else if (count == 0)
         return "";

      StringBuffer   result = new StringBuffer(count);

      for (int j = 0; j < count; ++j)
         result.append(c);

      return result.toString();
   }

   private static String padLeft(String s, char padChar, int finalLength)
   {
      if (s == null)
         return null;
      else if (s.length() >= finalLength)
         return s;

      return repeat(padChar, finalLength - s.length()) + s;
   }

   private static String right(String s, int length)
   {
      if (s == null)
         return null;
      else if (length < 0 && s.length() <= -length)
         return "";
      else if (s.length() <= length)
         return s;

      if (length < 0)
         return s.substring(0, s.length() + length);
      else
         return s.substring(s.length() - length);
   }

   public static byte[] truncate16(byte[] b, int offset, int length)
   {
     return extractBytes(b, offset, (length / 16) * 16);
   }

   /**
     This method will zero-pad its result to make sure it is 'length'
     bytes long, regardless of how many bytes are available from 'b'.
   */
   public static byte[] extractBytes(byte[] b, int offset, int length)
   {
     if (b == null)
       return null;

     byte[]  result = new byte[length];

     System.arraycopy(b, offset, result, 0, Math.min(length, b.length));

     return result;
   }
   
   public static int strlen(byte[] b)
   {
     if (b == null)
       return 0;

     int   len = 0;

     while (b[len] != 0)
       ++len;

     return len;
   }
}
